{
   muu TrueColor library for TMT Pascal - VESA/VBE Console driver
   Copyright (C) 1999 Tor Arne Gustad (anden@online.no)

   This source code is licensed under the GNU Public License.
   See the GNU General Public License for more details.
}

UNIT mc_VESA;

        // List of supported options:
        // -------------------
        // "NORMAL"                 -   use color fakemode
        // "GRAYSCALE"              -   use grayscale fakemode
        // "DITHERING"              -   use dithered fakemode++
        // "WIDTH <n>" "HEIGHT <n>" -   force width and height
        // "DEPTH <bpp>"            -   force Console depth
        // "FORCE BANKED MEMORY"    -   force banked mode only
        // "RETRACE CHECKING"       -   force video retrace checking
        // "PAGE FLIPPING"          -   enable page flipping support
        // "NO MESSAGES"            -   disable all output messages


INTERFACE
{$R-,I+,Q-}

    // WARNING: 
    //  Some of these options can be hazardous to your computer. They are 
    //  still under development and should not be changed if you don't know 
    //  exactly what you are doing:

//{$DEFINE SECOND_MODEDETECT} // non-standard modedetection routine (off)
//{$DEFINE PAGE_FLIP_SUPPORT} // enable page flipping support (off)
//{$DEFINE REMOVE_MTRR}     // toggle removing of MTRR upon VESA shutdown (off)

USES mArea,mClip,mFormat,mSurface,mConsole,VBE,VBECore,VBETypes,CPUInfo,MTRR,Detect,DPMI,STRINGS,Nova,HTimer,Memory,Error;
    // maximum number of offscreen video pages, default two
CONST pageflip_maxpages =  2;

    // VGA memory for Banked mode support (real_mode_address*10h)
CONST VGAmemory =   $0A0000;

// list of available modes --------------------------------------------------
TYPE VBEModeT     = RECORD
                      width,height: dword;
                      depth,mode:   word;
                    END;

TYPE VBEModeListT = RECORD
                      info: ARRAY[0..512] of VBEModeT;
                      max:  dword;
                    END;

// the console driver -------------------------------------------------------
TYPE VBEConsole =   ^VBEConsoleT;
     VBEConsoleT =  OBJECT (ConsoleT)
                     
                     public
                      CONSTRUCTOR Create;
                      DESTRUCTOR Destroy; VIRTUAL;

                      PROCEDURE Option(optv:pchar); VIRTUAL;

                        // initialization
                      PROCEDURE Open(title:pchar;lwidth,lheight:word;lformat:Format); VIRTUAL;
                      PROCEDURE Close; VIRTUAL;
                      FUNCTION Error:   boolean; VIRTUAL;

                        // update (banked mode flip, page flip update)
                      PROCEDURE Update; VIRTUAL;

                        // physical Console information
                      FUNCTION Info:    ConsoleInfo; VIRTUAL;

                        // indirect video memory interface
                      PROCEDURE Clear(m_area:Area;m_color:dword); VIRTUAL;
                      PROCEDURE Copy(tx,ty:longint;m_surface:Surface;m_area:Area); VIRTUAL;

                        // message processing
                      PROCEDURE Message(m_string:string); VIRTUAL;

                     private
                      vesa:     VESAInterface;  // the VESA interface object
                      modes:    VBEModeListT;   // list of available modes
                      clock:    Timer;
                      _error:   boolean;        // object status

                        // mode information
                      cformat:  Format;         // console pixel format
                      fakemode: byte;           // fake RGB mode
                      
                        // video physical information
                      framebuffer:  dword;      // the linear framebuffer
                      video_linear: boolean;    // linear or banked memory
                      video_size:   dword;      // size of one video page (B)

                        // page flipping variables
                      buffer_video: boolean;
                      buffer_data:  pointer;    // buffer data
                        // (video memory related)
                      buffer_pages: byte;       // number of pages
                      buffer_page:  byte;       // current page

                        // options
                      m_optwidth,m_optheight:   word;       // width, height
                      m_optformat:              Format;     // format
                      m_optforcebanked:         boolean;    // force banked
                      m_optpageflip:            boolean;    // use pageflip
                      m_retrace:                boolean;    // retrace check
                      m_displaymessages:        boolean;

                        // default options
                      m_default_width:          word;
                      m_default_height:         word;
                      m_default_format:         Format;

                        // special options
                      m_no_linear_bit:          boolean;
                      m_write_combine:          boolean;

                        // private VESA interface functions
                      PROCEDURE set_vesa_mode(w,h:word;d:Format);
                      PROCEDURE update_page_flip;
                      PROCEDURE sort_mode_list;

                    END;

IMPLEMENTATION

// --------------------------------------------------------------------------
CONSTRUCTOR VBEConsoleT.Create;
BEGIN
    // clear variables
  framebuffer := 0;
  video_size := 0;
  fakemode := 0; // set fake RGB to colors
  vesa := NIL;

    // set up defaults
  m_default_width := 320;
  m_default_height := 200;
  m_default_format := Format(BGR8888);

    // set up page flipping
  video_linear := false;
  buffer_video := false;
  buffer_data := NIL;
  buffer_pages := 1;
  buffer_page := 0;

    // clear Option variables
  m_optwidth := 0;
  m_optheight := 0;
  m_optformat := Format(Unknown);
  m_optforcebanked := false;
  m_optpageflip := false;
  m_retrace := false;
  m_displaymessages := true;

    // default special options
  m_no_linear_bit := false;
  m_write_combine := false;

    // open clock timer
  new(clock,Create);

    // open VESA interface
  new(vesa,open);

    // get VESA general information and update error flag
  vesa^.getinfo;
  _error := vesa^.error;
END;


DESTRUCTOR VBEConsoleT.Destroy;
  dispose(vesa,close);


PROCEDURE VBEConsoleT.Option;
  VAR upstr:    pointer := NIL;
      check:    integer := 0;
BEGIN
  upstr := strUpper(optv);
  
    // "normal" colors
  IF (strComp(upstr,console_option_normal) = 0) THEN
    fakemode := 0;

    // grayscale where available
  IF (strComp(upstr,console_option_grayscale) = 0) THEN
    fakemode := 1;

    // dither where available
  IF (strComp(upstr,console_option_dither) = 0) THEN
    fakemode := 2;

    // specify Console width
  IF (strLComp(upstr,console_option_width,length(console_option_width)) = 0) THEN
  BEGIN
    val(strPas(upstr+length(console_option_width)),m_optwidth,check);
    IF (check <> 0) THEN
      m_optwidth := 0;
  END;
  
    // specify Console height
  IF (strLComp(upstr,console_option_height,length(console_option_height)) = 0) THEN
  BEGIN
    val(strPas(upstr+length(console_option_height)),m_optheight,check);
    IF (check <> 0) THEN
      m_optheight := 0;
  END;

    // specify Console Format
  IF (strLComp(upstr,console_option_depth,length(console_option_depth)) = 0) THEN
    m_optformat := Format(strPas(upstr+length(console_option_depth)));

    // force linear modes only
  IF (strComp(upstr,console_option_banked) = 0) THEN
    m_optforcebanked := true;

    // use retrace checking (if not page flipping is used)
  IF (strComp(upstr,console_option_retrace) = 0) THEN
    m_retrace := true;

  {$IFDEF PAGE_FLIP_SUPPORT}
    // enable page flipping support
  IF (strComp(upstr,console_option_pageflip) = 0) THEN
    m_optpageflip := true;
  {$ENDIF}

    // turn off messages
  IF (strComp(upstr,console_option_nomessages) = 0) THEN
    m_displaymessages := false;
END;


PROCEDURE VBEConsoleT.Open;

        // this function checks for duplicate mode records
    FUNCTION FINDMODE(mode,width,height,depth:word): boolean;
      IF (modes.max > 0) THEN
      DECLARE VAR lpos: dword := 0;
      BEGIN
        FOR lpos := 0 TO (modes.max-1) DO
          IF ((modes.info[lpos].Width = width) AND
              (modes.info[lpos].Height = height) AND
              (modes.info[lpos].Depth = depth)) OR
             (modes.info[lpos].Mode = mode) THEN
            BEGIN
              result := true;
              exit;
            END ELSE
              result := false;
      END ELSE
        result := false;

  VAR m_str:    string;

BEGIN
    // check if an error has occurred while creating object
  IF (_error) THEN
    Exit;
    
  Message('Starting muuTC VESA Console driver...');

  video_linear := boolean(vesa^.vesainfo^.vbeversion >= $200);

  fillchar(modes,sizeof(modes),0);
  modes.max := 0;

  IF (NOT vesa^.error) THEN
  BEGIN

        // check for a valid mode list; exit if none found
    IF(memw[vesa^.vesainfo^.videomodePTR_SEG:vesa^.vesainfo^.videomodePTR_OFS] = $FFFF) THEN
    BEGIN
      _error := true;
      Exit;
    END;

    // This is the first attempt to get a list of available modes from VESA
    // implementation. It uses the modelistPTR found in the vesa information
    // record to extract modes supported by hardware.
    DECLARE VAR mpos,set_vesa_mode:dword := 0;
    BEGIN
      REPEAT
        set_vesa_mode := memw[vesa^.vesainfo^.videomodePTR_SEG*16+vesa^.vesainfo^.videomodePTR_OFS+mpos];
        vesa^.getmodeinfo(set_vesa_mode);

        IF (NOT vesa^.error) AND    // supported
           (vesa^.modeinfo^.bitsperpixel >= 8) AND  // depth
           boolean(vesa^.modeinfo^.modeattributes AND 16) AND   // graphical mode
           (NOT FINDMODE(set_vesa_mode,vesa^.modeinfo^.xresolution,vesa^.modeinfo^.yresolution,vesa^.modeinfo^.bitsperpixel)) THEN // duplicate
        BEGIN
          modes.info[modes.max].Mode := set_vesa_mode;
          modes.info[modes.max].Width := vesa^.modeinfo^.xresolution;
          modes.info[modes.max].Height := vesa^.modeinfo^.yresolution;
          modes.info[modes.max].Depth := vesa^.modeinfo^.bitsperpixel;
          modes.max +:= 1;
        END;
        mpos +:= 2;
      UNTIL (set_vesa_mode = $FFFF);
    END;

{$IFDEF SECOND_MODEDETECT}
    // The second attempt to get modes goes through the whole set of possible
    // modes (range 0x100 to 0x1000) and checks if VESA accepts it. Modes 
    // that already exists in the list are skipped.
    //  (This should NOT be needed in any case, but are here anyway)
    DECLARE VAR set_vesa_mode:word := 0;
    BEGIN
      FOR set_vesa_mode := $100 TO $1000 DO
      BEGIN
        vesa^.getmodeinfo(set_vesa_mode);

        IF (NOT vesa^.error) AND    // supported
           (vesa^.modeinfo^.bitsperpixel >= 8) AND  // depth
           boolean(vesa^.modeinfo^.modeattributes AND 16) AND   // graphical mode
           (NOT FINDMODE(set_vesa_mode,vesa^.modeinfo^.xresolution,vesa^.modeinfo^.yresolution,vesa^.modeinfo^.bitsperpixel)) THEN // duplicate
        BEGIN
          modes.info[modes.max].Mode := set_vesa_mode;
          modes.info[modes.max].Width := vesa^.modeinfo^.xresolution;
          modes.info[modes.max].Height := vesa^.modeinfo^.yresolution;
          modes.info[modes.max].Depth := vesa^.modeinfo^.bitsperpixel;
          modes.max +:= 1;
        END;
      END;
    END;
{$ENDIF}
  END;

    // check if any modes were found, exit if not
  IF (modes.max = 0) THEN
  BEGIN
    _error := true;
    Exit;
  END ELSE
    _error := false;

    // print VESA information++
  str('  VESA ',(vesa^.vesainfo^.vbeversion shr 8),'.',(vesa^.vesainfo^.vbeversion and $FF),
           ' with ',vesa^.vesainfo^.totalmemory*64,'Kb memory',
           ' and ',modes.max,' modes',
           ' detected',m_str);
  Message(m_str);  

    // sort the mode list [(low_res,high_depth)..(high_res,low_depth)]
  sort_mode_list;

    // clear some variables before mode initialization
  framebuffer := 0;
  video_size := 0;

    // set up defaults if none specified
  IF (lwidth = 0) OR
     (lheight = 0) THEN
  BEGIN
    lwidth := m_default_width;
    lheight := m_default_height;
  END;

  IF (lformat.pixelformat = Unknown) THEN
    lformat := Format(m_default_format);

    // use any Options specified
  IF (m_optwidth <> 0) THEN
    lwidth := m_optwidth;
  IF (m_optheight <> 0) THEN
    lheight := m_optheight;
  IF (m_optformat.pixelformat <> Unknown) THEN
    lformat := Format(m_optformat);
  IF (m_optforcebanked = true) THEN
  BEGIN
    Message('  forcing banked mode');
    lformat.Direct := false;
  END;

    // try to set the video mode, use special cases for some cards
  DECLARE VAR vendorname:   pchar := pchar(vesa^.vesainfo^.OEMStringPTR_SEG*16+vesa^.vesainfo^.OEMStringPTR_OFS);
              productname:  pchar := pchar(vesa^.vesainfo^.OEMProductNamePTR_SEG*16+vesa^.vesainfo^.OEMProductNamePTR_OFS);
  BEGIN

        // the Alliance AT25/AT3D does not approve of the linear mode bit
    IF ( (strLComp(strPos(vendorname,'AT3D'),'AT3D',4) = 0) OR
         (strLComp(strPos(productname,'AT3D'),'AT3D',4) = 0) ) THEN
    BEGIN
      Message('  using Alliance AT25/AT3D special setting');
      m_no_linear_bit := true;
    END;

    IF ( strLComp(strPos(vendorname,'VORTEX'),'VORTEX',6) = 0) THEN
    BEGIN
      Message('  overriding VORTEX VESA version to v1.2');
      vesa^.vesainfo^.vbeversion := $102;
    END;

    set_vesa_mode(lwidth,lheight,lformat);
  END;
END;

PROCEDURE VBEConsoleT.Close;
BEGIN
    // free physical video memory mapping
  IF (video_linear) AND (framebuffer <> 0) THEN
    freePhysicalMap(framebuffer);

    // free system memory page buffer
  IF (buffer_video = false) AND (NOT video_linear) THEN
    freemem(buffer_data,video_size);

{$IFDEF REMOVE_MTRR}
    // disable MTRR if installed
  IF (m_write_combine) THEN
  BEGIN
    mtrr_disable(vesa^.modeinfo^.physbaseptr,
      vesa^.vesainfo^.totalmemory*65536,mtrr_WriteCombine);
    m_write_combine := false;
  END;
{$ENDIF}
END;

FUNCTION VBEConsoleT.Error;
  result := _error; // return state




//  Clearing and copying
// --------------------------------------------------------------------------
    {$I include\palgen.pas}

PROCEDURE VBEConsoleT.Copy;
  VAR convert:  Nova_Converter := NIL;
BEGIN
  convert := Nova_getConverter(cpu_info,m_surface^.sFormat.pixelformat,cformat.pixelformat);

  IF (@convert = NIL) THEN
    fail('(mc_VESA) Unsupported color conversion.');

  DECLARE VAR put_width,put_height:     longint;
              scr_startx,scr_starty:    longint;
              data_start:               dword;
  BEGIN
    IF (tx < 0) THEN
      m_area.Left -:= tx;

    IF (ty < 0) THEN
      m_area.Top -:= ty;

    scr_startx := (tx+abs(tx)) shr 1;
    scr_starty := (ty+abs(ty)) shr 1;
    
    m_area := Clip(m_area,Area(0,0,m_surface^.Width,m_surface^.Height));
    m_area := Clip(m_area,Area(0,0,m_area.Left+vesa^.modeinfo^.xresolution-scr_startx,m_area.Top+vesa^.modeinfo^.yresolution-scr_starty));

    put_width := m_area.Right-m_area.Left;
    put_height := m_area.Bottom-m_area.Top;

    IF (put_width <= 0) OR
       (put_height <= 0) THEN
         Exit;

    data_start := (m_area.Left+m_area.Top*m_surface^.Width)*Pitch(m_surface);

        // do vertical retrace check if requested
    IF (m_retrace) AND (video_linear) THEN
      wait_for_vertical_retrace;

    IF (m_surface^.sFormat.Pixelformat = VPAL) THEN
      Nova_setVPal(m_surface^.sFormat.Palette);

    convert(ptr(ofs(m_surface^.Data^)+data_start),
            ptr(ofs(buffer_data^)+scr_startX*Pitch(cformat)+scr_startY*vesa^.modeinfo^.bytesperscanline),
            put_width,
            put_height,
            (m_surface^.Width-put_width)*Pitch(m_surface),
            vesa^.modeinfo^.bytesperscanline-put_width*Pitch(cformat));
  END; 
END;


PROCEDURE VBEConsoleT.Clear;
  VAR clearer:    Nova_Clearer := NIL;
BEGIN
  clearer := Nova_getClearer(cpu_info,Pitch(cformat));

  IF (@clearer = NIL) THEN
    fail('(mc_VESA) Unsupported clearer depth.');

  DECLARE VAR put_width,put_height: dword;
              data_start:           dword;
              pixel_clear:          dword;
  BEGIN
    m_area := Clip(m_area,Area(0,0,vesa^.modeinfo^.xresolution,vesa^.modeinfo^.yresolution));
    put_width := m_area.Right-m_area.Left;
    put_height := m_area.Bottom-m_area.Top;

    IF (put_width <= 0) OR
       (put_height <= 0) THEN
         Exit;

    data_start := (m_area.Top*vesa^.modeinfo^.bytesperscanline)+(m_area.Left*Pitch(cformat));

    CASE Depth(cformat) OF
      16:   pixel_clear := Pack16(m_color)+Pack16(m_color) shl 16;
      15:   pixel_clear := Pack15(m_color)+Pack15(m_color) shl 16;
       8:   pixel_clear := Pack8(m_color)+(Pack8(m_color) shl 8)+
                           (Pack8(m_color) shl 16)+(Pack8(m_color) shl 24);
     ELSE
       pixel_clear := m_color;

    END;

        // do vertical retrace check if requested
    IF (m_retrace) AND (video_linear) THEN
      wait_for_vertical_retrace;

    clearer(ptr(ofs(buffer_data^)+data_start),
            pixel_clear,
            put_width,
            put_height,
            vesa^.modeinfo^.bytesperscanline-put_width*Pitch(cformat));
  END; 
END;


PROCEDURE VBEConsoleT.Update;
  update_page_flip;


FUNCTION VBEConsoleT.Info;
BEGIN
  result.Width := vesa^.modeinfo^.xresolution;
  result.Height := vesa^.modeinfo^.yresolution;
  result.Format := Format(cformat);
  result.Name := 'mtc VESA Console driver';
  result.Area := Area(0,0,vesa^.modeinfo^.xresolution,vesa^.modeinfo^.yresolution);
END;


PROCEDURE VBEConsoleT.Message;
  IF (m_displaymessages) THEN
  DECLARE VAR m_text:   boolean := true;
  BEGIN
    ASM
        // check if we are in text mode before writing
      mov ah, $0F
      int $10

      cmp al, $03
      je  @@textmode

      mov m_text,  0

    @@textmode:
    END;
    
    IF (m_text) THEN
      writeln(m_string);
  END;


//  Internal VESA Console functions
// --------------------------------------------------------------------------
PROCEDURE VBEConsoleT.set_vesa_mode;  // set mode
  VAR found_mode:   word := $FFFF;

    // This function locates a mode inside the modelist which is aquired in
    // the object constructor. It returns the first mode match in list.
    FUNCTION findmode(width,height,depth:word): word;
      IF (modes.max > 0) THEN
        DECLARE VAR lpos: dword := 0;
        BEGIN
          FOR lpos := 0 TO (modes.max-1) DO
            IF (modes.info[lpos].Width = width) AND
               (modes.info[lpos].Height = height) AND
               (modes.info[lpos].Depth = depth) THEN
              BEGIN
                result := lpos;
                Exit;
              END ELSE
                result := $FFFF;
        END ELSE
          result := $FFFF;

  VAR m_str:    string;
      m_rch:    boolean := false;
      m_linear: boolean := d.Direct;

BEGIN
    // first check if the wanted mode is available
  found_mode := findmode(w,h,Depth(d));

  IF (found_mode <> $FFFF) THEN
    vesa^.getmodeinfo(modes.info[found_mode].mode)
  ELSE
    vesa^.seterror(true);
 
    // if VESA failed or no mode was found, try to locate another one
    // (since the list now is sorted, this code will result in always setting
    //  the best supported mode available)
    IF (vesa^.error) THEN
    BEGIN
      str('  unable to set ',w,'x',h,' @ ',Depth(d),'bpp',m_str);
      Message(m_str);

        // start with best available format
      d := Format(32);

        // try to locate lower depth, same res.
      REPEAT
        found_mode := 0;

        WHILE (found_mode <= modes.max) DO
        BEGIN
          vesa^.getmodeinfo(modes.info[found_mode].mode);

          IF ( (NOT vesa^.error) AND
               (vesa^.modeinfo^.xresolution = w) AND
               (vesa^.modeinfo^.yresolution = h) AND
               (vesa^.modeinfo^.bitsperpixel > Depth(d)) ) THEN
            BEGIN
              m_rch := true;
              Break;
            END;

          found_mode +:= 1;
        END;

        IF ( Depth(d) = 8 ) THEN
          d := Format(Unknown);
        IF ( Depth(d) = 15 ) THEN
          d := Format(8);
        IF ( Depth(d) = 16 ) THEN
          d := Format(15);
        IF ( Depth(d) = 24 ) THEN
          d := Format(16);
        IF ( Depth(d) = 32 ) THEN
          d := Format(24);

      UNTIL ( d.PixelFormat = Unknown ) OR ( m_rch );

        // try to locate lower depth, same or higher res.
      IF (NOT m_rch) THEN
      BEGIN
        d := Format(32);

        REPEAT
          found_mode := 0;

            // try until VESA says ok or there are no modes left to try
          WHILE (found_mode <= modes.max) DO
          BEGIN
            vesa^.getmodeinfo(modes.info[found_mode].mode);

            IF ( (NOT vesa^.error) AND
                 (vesa^.modeinfo^.xresolution >= w) AND
                 (vesa^.modeinfo^.yresolution >= h) AND
                 (vesa^.modeinfo^.bitsperpixel > Depth(d)) ) THEN
              BEGIN
                m_rch := true;
                Break;
              END;

            found_mode +:= 1;
          END;

          IF ( Depth(d) = 8 ) THEN
            d := Format(Unknown);
          IF ( Depth(d) = 15 ) THEN
            d := Format(8);
          IF ( Depth(d) = 16 ) THEN
            d := Format(15);
          IF ( Depth(d) = 24 ) THEN
            d := Format(16);
          IF ( Depth(d) = 32 ) THEN
            d := Format(24);

        UNTIL ( d.PixelFormat = Unknown ) OR ( m_rch );

      END;

      IF (m_rch) THEN
      BEGIN
        str('    found mode ',vesa^.modeinfo^.xresolution,'x',vesa^.modeinfo^.yresolution,' @ ',vesa^.modeinfo^.bitsperpixel,'bpp',m_str);
        Message(m_str);
        vesa^.seterror(false);
      END;
    END;

  _error := vesa^.error;
    
    // exit if no modes were found
  IF (_error) THEN
    Exit;

    // convert the 32k direct color modes of VBE1.2+ BIOS'es to 15bpp if the
    // alpha channel is on (16bpp-Alpha)=15bpp
  IF (vesa^.modeinfo^.bitsperpixel = 16) AND
     (vesa^.modeinfo^.rsvdmasksize = 1) THEN
    vesa^.modeinfo^.bitsperpixel := 15;

    // set up correct Console pixel format for interface with Nova converter
  IF (vesa^.modeinfo^.redfieldposition > vesa^.modeinfo^.bluefieldposition) THEN
    CASE (vesa^.modeinfo^.bitsperpixel) OF
      32:   cformat := Format(BGR8888);
      24:   cformat := Format(BGR888);
      16:   CASE (fakemode) OF
              0,1:    cformat := Format(BGR565)
              2:      cformat := Format(DITHER16);
            END;
      15:   CASE (fakemode) OF
              0,1:    cformat := Format(BGR555)
              2:      cformat := Format(DITHER15);
            END;
    END ELSE    // R0GxBx DAC (I've never seen one...)
    CASE (vesa^.modeinfo^.bitsperpixel) OF
      32:   cformat := Format(RGB8888);
      24:   cformat := Format(RGB888);
      16:   cformat := Format(RGB565)
      15:   cformat := Format(RGB555);
       8:   CASE (fakemode) OF
              0:  cformat := Format(RGB323);
              1:  cformat := Format(GRAY8);
              2:  cformat := Format(DITHER8);
            END;
    END;

    // get screen size (size of one page)
  video_size := vesa^.modeinfo^.bytesperscanline*vesa^.modeinfo^.yresolution;

{$IFDEF PAGE_FLIP_SUPPORT}
  buffer_pages := pageflip_maxpages;
{$ELSE}
  buffer_pages := 1;
{$ENDIF}

    // restore direct color setting
  cformat.Direct := m_linear;
  d.Direct := m_linear;

    // check if there is enough video memory for the mode
  DECLARE VAR video_memory: dword := vesa^.vesainfo^.totalmemory*65536;
    WHILE (video_memory < (video_size*buffer_pages)) DO
      buffer_pages -:= 1;

    // exit if short on video memory
  IF (buffer_pages = 0) THEN
  BEGIN
    _error := true;
    Exit;
  END;

  IF (buffer_pages > 1) THEN
    m_retrace := false;

    // Linear Framebuffer initialization
  IF (vesa^.modeinfo^.physbaseptr <> 0) AND
     (d.Direct) THEN
  BEGIN
    framebuffer := 0;  // clear LFB status

        // decide whether to buffer in video memory or not
    IF (buffer_pages > 1) THEN
      buffer_video := true
    ELSE
      buffer_video := false;

        // map physical memory offset
    framebuffer := mapphysicaltolinear(vesa^.modeinfo^.physbaseptr,video_size*buffer_pages);

    str('  using linear framebuffer @ ',vesa^.modeinfo^.physbaseptr div (1024*1024),'MB',m_str);
    Message(m_str);

        // enable MTRR if the following conditions are present:
        //  - CPU supports MTRR and MSR read/write
        //  - EMS memory is not installed (e.g EMM386, QEMM)
        //  - Program is running in first priority (e.g not under Windows)
    IF (cpu_info.MTRR) AND (cpu_info.MSR) AND
       (EMSVersion = -1) AND (WINVersion = -1) THEN
    BEGIN
      str('    enabling MTRR Write-Combining',m_str);
      Message(m_str);
      mtrr_enable(vesa^.modeinfo^.physbaseptr,
                  vesa^.vesainfo^.totalmemory*65536,mtrr_WriteCombine);
      m_write_combine := true;
    END;

        // here we decide whether to use LFB or not
    video_linear := boolean(framebuffer <> 0);
  END ELSE BEGIN
    video_linear := false;  // Banked mode
    buffer_video := false;  //  ...
  END;

  // set up page buffering
  IF (buffer_video) THEN
  BEGIN
    buffer_page := 1;
    buffer_data := ptr(framebuffer+video_size);
    str('  page buffering enabled with ',buffer_pages,' pages',m_str);
    Message(m_str);
  END ELSE IF (video_linear) THEN
    buffer_data := ptr(framebuffer)
  ELSE BEGIN
    getmem(buffer_data,video_size);
    memset(buffer_data,video_size,0);
  END;

  str('  setting mode ',vesa^.modeinfo^.xresolution,'x',vesa^.modeinfo^.yresolution,' @ ',vesa^.modeinfo^.bitsperpixel,'bpp',m_str);
  Message(m_str);

    // Wait 0.6seconds
  clock^.Wait(600000);

    // finally, try to set the video mode
  DECLARE VAR mode: word := modes.info[found_mode].mode;
  BEGIN
        // set linear bit (d14) if this should be used
    mode := mode OR ($4000*(BYTE(video_linear)*BYTE(NOT m_no_linear_bit)));
    vesa^.setmode(mode);
    _error := vesa^.error;
  END;

  // exit in case of an error during "setmode"
  IF (_error) THEN
    Exit;

  // set the correct palette if an 8bit mode has been set
  IF (Depth(cformat) = 8) THEN
  DECLARE VAR pal:  VBE_VESAPalette;
              dacw: byte := 0;
  BEGIN
    new(pal);

    // try to get DAC width from VESA, fallback to 6bit on error
    //  (bugfix by pothed)
    dacw := vesa^.getDACWidth;
    IF (vesa^.error) THEN
      dacw := 6;

        // create appropriate palette for mode
    CASE (fakemode) OF
      0:  createpal_rgb8(pal,(1 shl dacw)-1);
      1:  createpal_gray8(pal,(1 shl dacw)-1);
      2:  createpal_dither(pal,(1 shl dacw)-1);
    END;

        // set the palette through VESA(LFB) or VGA(Banked)
    CASE (video_linear) OF
      true:   vesa^.setpalette(pal,0,256);
      false:  vga_setpalette(pal,0,256);
    END;

    dispose(pal);
  END;
 
    // exit with error if anything failed
  IF (vesa^.error) THEN
  BEGIN
    _error := vesa^.error;
    Exit;
  END;
END;


PROCEDURE VBEConsoleT.update_page_flip;
  IF (buffer_video) THEN    // Page flipping
  BEGIN
    vesa^.setdisplaystart(0,buffer_page*vesa^.modeinfo^.yresolution);
    buffer_page +:= 1;
    IF (buffer_page >= buffer_pages) THEN
      buffer_page := 0;
    
    buffer_data := ptr(framebuffer+buffer_page*video_size);
  END ELSE IF (NOT video_linear) THEN   // Banked mode flip (realmode)
  DECLARE VAR bank:         byte := 0;
              bank_size:    dword := vesa^.modeinfo^.winsize*1024;
              bank_next:    byte := (bank_size div (vesa^.modeinfo^.wingranularity*1024));
              buffer_pos:   dword := 0;
              buffer_move:  dword := 0;
  BEGIN
    IF (m_retrace) THEN
      wait_for_vertical_retrace;
    REPEAT
          // set VESA window
      vesa^.setwindow(WindowA,bank);
 
          // check how much memory to move (bytes)
      IF ((video_size-buffer_pos) < bank_size) THEN
        buffer_move := (video_size-buffer_pos)
      ELSE
        buffer_move := bank_size;

          // copy Console buffer data to screen
      memcpy(ptr(ofs(buffer_data^)+buffer_pos),ptr(VGAmemory),buffer_move);

          // next bank, next buffer position
      bank +:= bank_next;
      buffer_pos +:= buffer_move;
    UNTIL (buffer_pos >= video_size);
  END;


PROCEDURE VBEConsoleT.sort_mode_list;

    // sorts [res,depth] in the following order:
    //
    // [low,low] (..) [high,high]

    VAR c1,c2:  longint := 0;
        mode:   VBEmodeT;
BEGIN
  IF( modes.max = 0 ) THEN
    Exit;

    // sort resolution [low..high]
  FOR c1 := 0 TO (modes.max-1) DO
    FOR c2 := c1 TO (modes.max-1) DO
      IF ((modes.info[c1].width*modes.info[c1].height) > (modes.info[c2].width*modes.info[c2].height)) THEN
      BEGIN
        move(modes.info[c1],mode,sizeof(VBEmodeT));
        move(modes.info[c2],modes.info[c1],sizeof(VBEmodeT));
        move(mode,modes.info[c2],sizeof(VBEmodeT));
      END;

    // sort depth [high..low] (but do not disturb resolution sorting...:)
  FOR c1 := 0 TO (modes.max-1) DO
    FOR c2 := c1 TO (modes.max-1) DO
      IF ((modes.info[c1].depth) > (modes.info[c2].depth)) AND
         ((modes.info[c1].width*modes.info[c1].height) >= (modes.info[c2].width*modes.info[c2].height)) THEN
      BEGIN
        move(modes.info[c1],mode,sizeof(VBEmodeT));
        move(modes.info[c2],modes.info[c1],sizeof(VBEmodeT));
        move(mode,modes.info[c2],sizeof(VBEmodeT));
      END;
END;

end.
